use std::collections::HashMap;

use crate::{OpCode, LispValue};

pub struct LispState {
	table: HashMap<String, LispValue>,
}

fn add_function<T: Into<String> + Clone>(table: &mut HashMap<String, LispValue>, name: T, func: fn(Vec<LispValue>) -> Vec<LispValue>) {
	table.insert(name.clone().into(), LispValue::RustFunction(name.into(), func));
}

impl LispState {
	pub fn new() -> LispState {
		let mut table = HashMap::new();

		add_function(&mut table, "print", |x| {
			let mut strings = Vec::new();
			for val in x {
				strings.push(val.to_string());
			}

			let str = strings.join(" ");
			println!("{}", str);

			vec![LispValue::Nil]
		});

		add_function(&mut table, "add", |x| {
			if x.len() != 2 {
				return vec![LispValue::Nil]
			}

			if let (LispValue::Integer(lhs), LispValue::Integer(rhs)) = (&x[0], &x[1]) {
				return vec![LispValue::Integer(lhs + rhs)];
			} else if let (LispValue::String(lhs), LispValue::String(rhs)) = (&x[0], &x[1]) {
				return vec![LispValue::String(lhs.to_owned() + rhs)];
			} else {
				return vec![LispValue::Nil];
			}
		});

		LispState {
			table
		}
	}

	fn do_call(&self, opcode: &OpCode) -> Vec<LispValue> {
		if let OpCode::Call(func, args) = opcode {
			let f = self.table.get(func).unwrap();

			let mut args: Vec<LispValue> = args.clone();
			for i in 0..args.len() {
				let arg = &args[i];
				if let LispValue::Eval(OpCode::Exp(opcodes)) = arg {
					if opcodes.len() < 1 {
						args[i] = LispValue::Nil;
						continue;
					}

					let call_opcode = opcodes.last().unwrap();
					let mut res = self.do_call(call_opcode);
					args.remove(i);
					for j in 0..res.len() {
						args.insert(i + j, res.remove(0));
					}
				}
			}
			if let LispValue::RustFunction(_, f) = f {
				return f(args.to_vec());
			} else {
				todo!();
			}

		} else {
			panic!("Bad OpCode: expected OpCode::Call")
		}
	}

	pub fn execute(&self, instructions: Vec<OpCode>) {
		for op in instructions {
			match op {
				OpCode::Call(func, args) => {
					self.do_call(&OpCode::Call(func, args));
				},
				OpCode::Exp(ins) =>  {
					self.execute(ins);
				}
			}
		}
	}
}
